#!/usr/local/bin/perl

use strict;
use Encode qw/encode decode/;
use utf8;

my $codepage;
my @rttbl;
my $count = 0;

sub str2hex {
	my $str = shift;
	my $c1;

	$c1 = unpack('C', $str);
	if (0x00 <= $c1 && $c1 <= 0x80 || $c1 == 0xFF) {
		sprintf("%02X", $c1);
	} else {
		sprintf("%02X%02X", unpack('CC', $str));
	}
}

sub usage {
	print "Usage: perl cprttest.pl codepage\n";
	exit;
}

if ($#ARGV < 0) { &usage(); }

if ($ARGV[0] =~ /^(cp\d+)/i) {
	$codepage = $1;
} elsif ($ARGV[0] =~ /^(\d+)/i) {
	$codepage = "cp$1";
} else {
	&usage();
}

if (!open(FILE, "${codepage}rttbl.txt")) {
	die print "Can not open file(${codepage}rttbl.txt).\n";
}

while (<FILE>) {
	/^0x([0-9A-F]+).*U\+([0-9A-F]+).*0x([0-9A-F]+)/io or next;
	my $src = pack('CC', (hex($1) >> 8) & 0xFF, hex($1) & 0xFF);
	my $dst = pack('CC', (hex($3) >> 8) & 0xFF, hex($3) & 0xFF);
	my $uni = hex($2);

	push (@rttbl, [$src, $uni, $dst]);
}
close(FILE);

for (my $i = 0; $i <= $#rttbl; $i++) {
	my $src = $rttbl[$i][0];
	my $uni = $rttbl[$i][1];
	my $dst = $rttbl[$i][2];
	my $utf8 = decode($codepage, $src);
	my $out  = encode($codepage, $utf8);

	if ($dst ne $out) {
		my $srchex = str2hex($src);
		my $dsthex = str2hex($dst);
		my $outhex = str2hex($out);
		printf( "INPUT=0x%s -> UNICODE=U+%04X -> RESULT=0x%s\t(MS=0x%s(U+%04X))\n",
			$srchex, ord($utf8), $outhex, $dsthex, $uni );
		$count++;
	}
}
exit($count);
