#include <windows.h>
#include <winnls.h>
#include <stdio.h>
#include <locale.h>

typedef struct {
	int start;
	int end;
} CODERANGE;

typedef struct {
	int        CodePage;
	int        NumOfSBCS;
	int        NumOfDBCS1st;
	int        NumOfDBCS2nd;
	CODERANGE *SBCSRange;
	CODERANGE *DBCS1stRange;
	CODERANGE *DBCS2ndRange;
} CODEPAGEINFO;

/*-- Code Page 932 (SHIFT-JIS:Japanese) --*/
CODERANGE SBCS932[] = {
	{ 0x00, 0x7F },
	{ 0xA1, 0xDF }      /* JIS X 0201 Katakana */
};
CODERANGE DBCS932_1st[] = {
	{ 0x81, 0x9F },
	{ 0xE0, 0xFC }
};
CODERANGE DBCS932_2nd[] = {
	{ 0x40, 0x7E },
	{ 0x80, 0xFC }
};

/*-- Code Page 936 (GBK:Simplified Chinese) --*/
CODERANGE SBCS936[] = {
	{ 0x00, 0x80 }      /* 0x80 = EURO SIGN */
};
CODERANGE DBCS936_1st[] = {
	{ 0x81, 0xFE }
};
CODERANGE DBCS936_2nd[] = {
	{ 0x40, 0x7E },
	{ 0x80, 0xFE }
};

/*-- Code Page 949 (UHC:Korean) --*/
CODERANGE SBCS949[] = {
	{ 0x00, 0x7F }
};
CODERANGE DBCS949_1st[] = {
	{ 0x81, 0xFE }
};
CODERANGE DBCS949_2nd[] = {
	{ 0x41, 0x5A },     /* 'A' - 'Z' */
	{ 0x61, 0x7A },     /* 'a' - 'z' */
	{ 0x81, 0xFE }
};

/* -- Code Page 950 (BIG5:Traditional Chinese) --*/
CODERANGE SBCS950[] = {
	{ 0x00, 0x7F }
};
CODERANGE DBCS950_1st[] = {
	{ 0x81, 0xFE }
};
CODERANGE DBCS950_2nd[] = {
	{ 0x40, 0x7E },
	{ 0xA1, 0xFE }
};

CODEPAGEINFO CpInfo[] = {
	{
		932,
		sizeof(SBCS932)     / sizeof(CODERANGE),
		sizeof(DBCS932_1st) / sizeof(CODERANGE),
		sizeof(DBCS932_2nd) / sizeof(CODERANGE),
		SBCS932,
		DBCS932_1st,
		DBCS932_2nd
	},
	{
		936,
		sizeof(SBCS936)     / sizeof(CODERANGE),
		sizeof(DBCS936_1st) / sizeof(CODERANGE),
		sizeof(DBCS936_2nd) / sizeof(CODERANGE),
		SBCS936,
		DBCS936_1st,
		DBCS936_2nd
	},
	{
		949,
		sizeof(SBCS949)     / sizeof(CODERANGE),
		sizeof(DBCS949_1st) / sizeof(CODERANGE),
		sizeof(DBCS949_2nd) / sizeof(CODERANGE),
		SBCS949,
		DBCS949_1st,
		DBCS949_2nd
	},
	{
		950,
		sizeof(SBCS950)     / sizeof(CODERANGE),
		sizeof(DBCS950_1st) / sizeof(CODERANGE),
		sizeof(DBCS950_2nd) / sizeof(CODERANGE),
		SBCS950,
		DBCS950_1st,
		DBCS950_2nd
	}
};

mb2wc( int CodePage, wchar_t *pwc, BYTE *pmb )
{
	wchar_t wcstr[2];
	int num;

	num = MultiByteToWideChar(
	                    CodePage,
	                    MB_ERR_INVALID_CHARS,
	                    pmb,
	                    -1,
	                    wcstr,
	                    sizeof(wcstr)/sizeof(wchar_t)+1 );
	if (num) {
		*pwc = wcstr[0];
	}
	return num;
}

wc2mb( int CodePage, BYTE *pmb, wchar_t wc )
{
	wchar_t wcstr[2];
	BYTE mbstr[4];
	BOOL UsedDefaultChar = FALSE;
	int num;

	wcstr[0] = wc;
	wcstr[1] = L'\0';

	num = WideCharToMultiByte(
	                     CodePage,
	                     0,
	                     (LPCWSTR)wcstr,
	                     -1,
	                     (LPSTR)mbstr,
	                     sizeof(mbstr)+1,
	                     NULL,
	                     &UsedDefaultChar );
	if (num) {
		strncpy(pmb, mbstr, strlen(mbstr) + 1);
	}
	return UsedDefaultChar;
}

int roundtrip( int CodePage, BYTE *pmbsrc, wchar_t *pwc, BYTE *pmbdst)
{
	int ret = 0;

	if (mb2wc( CodePage, pwc, pmbsrc )) {
		if (!wc2mb( CodePage, pmbdst, *pwc )) {
			ret = 1;
		}
	}
	return ret;
}

void write_rttbl( FILE *fp, int CodePage, BYTE *pmbsrc, wchar_t wc, BYTE *pmbdst)
{
	char direction[10];

	strcpy( direction, strcmp( pmbsrc, pmbdst ) ? "-->" : "<->" );
	if (strlen(pmbsrc) == 2) {
			fprintf( fp, "0x%02X%02X %s U+%04X %s 0x%02X%02X\n",
			pmbsrc[0], pmbsrc[1],
			direction,
			wc,
			direction,
			pmbdst[0], pmbdst[1]);
	}
	else {
			fprintf( fp, "0x%02X %s U+%04X %s 0x%02X\n",
			pmbsrc[0],
			direction,
			wc,
			direction,
			pmbdst[0]);
	}
}

void main(void)
{
	FILE *fprt;
	char rttblfile[256];
	BYTE mbsrc[10];
	BYTE mbdst[10];
	wchar_t wc;
	int i, j, k;
	int c1, c2;

	for (i = 0; i < sizeof(CpInfo)/sizeof(CODEPAGEINFO); i++) {
		sprintf( rttblfile, "cp%drttbl.txt", CpInfo[i].CodePage );

		if ((fprt = fopen( rttblfile, "w" )) == NULL) {
			printf( "Can not open file(%s).\n", rttblfile );
			exit(1);
		}
		printf( "Code Page %d\n", CpInfo[i].CodePage );

		for (j = 0; j < CpInfo[i].NumOfSBCS; j++) {
			for (c1 =  CpInfo[i].SBCSRange[j].start;
			     c1 <= CpInfo[i].SBCSRange[j].end;
			     c1++)
			{
				mbsrc[0] = (BYTE)c1;
				mbsrc[1] = '\0';
				mbdst[0] = mbdst[1] = mbdst[2] = '\0';
				wc = L'\0';

				if (roundtrip( CpInfo[i].CodePage, mbsrc, &wc, mbdst )) {
					write_rttbl( fprt, CpInfo[i].CodePage, mbsrc, wc, mbdst );
				}
			}
		}
		for (j = 0; j < CpInfo[i].NumOfDBCS1st; j++) {
			for (c1 =  CpInfo[i].DBCS1stRange[j].start;
			     c1 <= CpInfo[i].DBCS1stRange[j].end;
			     c1++)
			{
				for (k = 0; k < CpInfo[i].NumOfDBCS2nd; k++) {
					for (c2 =  CpInfo[i].DBCS2ndRange[k].start;
					     c2 <= CpInfo[i].DBCS2ndRange[k].end;
					     c2++)
					{
						mbsrc[0] = (BYTE)c1;
						mbsrc[1] = (BYTE)c2;
						mbsrc[2] = '\0';
						mbdst[0] = mbdst[1] = mbdst[2] = '\0';
						wc = L'\0';

						if (roundtrip( CpInfo[i].CodePage, mbsrc, &wc, mbdst )) {
							write_rttbl( fprt, CpInfo[i].CodePage, mbsrc, wc, mbdst );
						}
					}
				}
			}
		}
		fclose(fprt);
	}
}
